use super::object::Object;
use super::PlainText;
use super::StandardProperties;
use super::Timestamp;

#[derive(Debug)]
pub struct Paragraph<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug)]
pub struct Comment<'s> {
    pub source: &'s str,
    pub value: Vec<&'s str>,
}

#[derive(Debug)]
pub struct TableCell<'s> {
    pub source: &'s str,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug)]
pub struct VerseBlock<'s> {
    pub source: &'s str,
    pub name: Option<&'s str>,
    pub data: Option<&'s str>,
    pub children: Vec<Object<'s>>,
}

#[derive(Debug)]
pub struct CommentBlock<'s> {
    pub source: &'s str,
    pub name: Option<&'s str>,
    pub contents: &'s str,
}

pub type CharOffsetInLine = u16;

#[derive(Debug)]
pub enum RetainLabels {
    No,
    Yes,
    /// When adding -k to the switches on an example or src block, the labels are kept in the source code and links will use line numbers.
    Keep(CharOffsetInLine),
}

#[derive(Debug)]
pub struct ExampleBlock<'s> {
    pub source: &'s str,
    pub name: Option<&'s str>,
    pub switches: Option<&'s str>,
    pub number_lines: Option<SwitchNumberLines>,
    pub preserve_indent: Option<CharOffsetInLine>,
    pub retain_labels: RetainLabels,
    pub use_labels: bool,
    pub label_format: Option<&'s str>,
    pub contents: String,
}

#[derive(Debug)]
pub struct ExportBlock<'s> {
    pub source: &'s str,
    pub name: Option<&'s str>,
    pub data: Option<&'s str>,
    pub contents: &'s str,
}

#[derive(Debug)]
pub struct SrcBlock<'s> {
    pub source: &'s str,
    pub name: Option<&'s str>,
    pub language: Option<&'s str>,
    pub switches: Option<&'s str>,
    pub parameters: Option<&'s str>,
    pub number_lines: Option<SwitchNumberLines>,
    pub preserve_indent: Option<CharOffsetInLine>,
    pub retain_labels: RetainLabels,
    pub use_labels: bool,
    pub label_format: Option<&'s str>,
    pub contents: String,
}

#[derive(Debug)]
pub struct Clock<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct DiarySexp<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct Planning<'s> {
    pub source: &'s str,
    pub scheduled: Option<Timestamp<'s>>,
    pub deadline: Option<Timestamp<'s>>,
    pub closed: Option<Timestamp<'s>>,
}

#[derive(Debug)]
pub struct FixedWidthArea<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct HorizontalRule<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct Keyword<'s> {
    pub source: &'s str,
    pub key: &'s str,
    pub value: &'s str,
}

#[derive(Debug)]
pub struct BabelCall<'s> {
    pub source: &'s str,
    pub key: &'s str,
    pub value: &'s str,
}

#[derive(Debug)]
pub struct LatexEnvironment<'s> {
    pub source: &'s str,
}

/// A line number used in switches to lesser blocks.
///
/// This must be signed because emacs subtracts 1 from the actual value in the org-mode text, which makes a 0 turn into a -1.
pub type LineNumber = isize;

#[derive(Debug)]
pub enum SwitchNumberLines {
    New(LineNumber),
    Continued(LineNumber),
}

impl<'s> Paragraph<'s> {
    pub(crate) fn of_text(input: &'s str) -> Self {
        let mut objects = Vec::with_capacity(1);
        objects.push(Object::PlainText(PlainText { source: input }));
        Paragraph {
            source: input,
            children: objects,
        }
    }
}

impl<'s> StandardProperties<'s> for Paragraph<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for TableCell<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Comment<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for VerseBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}
impl<'s> StandardProperties<'s> for CommentBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}
impl<'s> StandardProperties<'s> for ExampleBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}
impl<'s> StandardProperties<'s> for ExportBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}
impl<'s> StandardProperties<'s> for SrcBlock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Clock<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for DiarySexp<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Planning<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for FixedWidthArea<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for HorizontalRule<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for Keyword<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for BabelCall<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> StandardProperties<'s> for LatexEnvironment<'s> {
    fn get_source<'b>(&'b self) -> &'s str {
        self.source
    }
}

impl<'s> Comment<'s> {
    pub fn get_value(&self) -> String {
        // TODO: maybe we should handle parsing here instead of storing the parsing result in the AST since I imagine getting the value of comments won't be a common operation.
        let final_size = self.value.iter().map(|line| line.len()).sum();
        let mut ret = String::with_capacity(final_size);
        for line in &self.value {
            ret.push_str(line);
        }

        ret
    }
}
